//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "Taper.h"
#include "StringHelp.h"
#include <clipbrd.hpp>
#include <IniFiles.hpp>
#include <math.h>
#include <shellapi.h>
#include "MathsHelp.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TTaperForm *TaperForm;
//---------------------------------------------------------------------------
__fastcall TTaperForm::TTaperForm(TComponent* Owner)
	: TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TTaperForm::FormCreate(TObject *Sender)
{

	//set up tapers array with MT3 only
	Tapers.Length = 1;
	Tapers[0].taperlength = 84;
	Tapers[0].smalldia = 19.86019;
	Tapers[0].largedia = 24.07615;
	Tapers[0].Name = "MT3 Taper Definition File Missing";

	//read taper definitions
	AnsiString FileName = ExtractFileDir(Application->ExeName) + "\\Tapers.csv";
	if (FileExists(FileName)) {
		TStringList *inputfile = new TStringList();
		inputfile->LoadFromFile(FileName);

        for (int i = 0; i < inputfile->Count; i++) {
            if (LeftString(inputfile->Strings[i],1) == "#") {
				inputfile->Delete(i);
				i--;
				}
			}
		Tapers.Length = inputfile->Count;
		String text, segment;
		double value;
		bool mm;
		double scale;

		for (int i = 0; i < inputfile->Count; i++) {
            text = inputfile->Strings[i];
            text = text + ",";
			//Extract the Taper Name
			Tapers[i].Name = ExtractWord(text,",");
			text = DeleteWord(text,",");
			segment = ExtractWord(text,",");
			mm = (segment.LowerCase() == "mm");
			//internally tapers are always held in mm
			if (mm) {
				scale = 1.0;
				} else {
				scale = 25.4;
				}
			text = DeleteWord(text,",");
			value = ToDoubleDef(ExtractWord(text,","),84);
			Tapers[i].taperlength = value * scale;
			text = DeleteWord(text,",");
			value = ToDoubleDef(ExtractWord(text,","),19.86019);
			Tapers[i].smalldia = value * scale;
			text = DeleteWord(text,",");
			value = ToDoubleDef(ExtractWord(text,","),24.07615);
			Tapers[i].largedia = value * scale;
			//text = DeleteWord(text,",");
            //the comment is discarded
			}

		delete inputfile;
		}


	//fill the TaperComboBox drop down menu
	TaperComboBox->Items->Clear();
	for (int i = 0; i < Tapers.Length; i++) {
		TaperComboBox->Items->Add(Tapers[i].Name);
		}


	//Read settings & put them here
	TIniFile *IniFile = new TIniFile(ExtractFileDir(Application->ExeName) + "\\Data.ini");

	NumberLines->Checked = IniFile->ReadBool("LatheTaperExt","NumberLines",false);
	LineNumberIncrement->Text = IniFile->ReadString("LatheTaperExt","LineNumberIncrement","10");
	Units->ItemIndex = IniFile->ReadInteger("LatheTaperExt","Units",0);
	HideComments->Checked = IniFile->ReadBool("LatheTaperExt","HideComments",false);
	MeasurementMode->ItemIndex = IniFile->ReadInteger("LatheTaperExt","MeasurementMode",0);
	FeedRateMode->ItemIndex = IniFile->ReadInteger("LatheTaperExt","FeedRateMode",0);
	TaperComboBox->ItemIndex = IniFile->ReadInteger("LatheTaperExt","TaperSize",0);
	Waypoint->Checked = IniFile->ReadBool("LatheTaperExt","Waypoint",false);
	WaypointX->Text = IniFile->ReadString("LatheTaperExt","WaypointX","0.0");
	WaypointZ->Text = IniFile->ReadString("LatheTaperExt","WaypointZ","0.0");
	StartpointX->Text = IniFile->ReadString("LatheTaperExt","StartpointX","0.0");
	StartpointZ->Text = IniFile->ReadString("LatheTaperExt","StartpointZ","0.0");
	//EndpointX->Text = IniFile->ReadString("LatheTaperExt","EndpointX","0.0");
	//EndpointZ->Text = IniFile->ReadString("LatheTaperExt","EndpointZ","0.0");
	ClearanceX->Text = IniFile->ReadString("LatheTaperExt","ClearanceX","1.0");
	ClearanceZ->Text = IniFile->ReadString("LatheTaperExt","ClearanceZ","1.0");
	Faced->Checked = IniFile->ReadBool("LatheTaperExt","Faced",false);
	FaceZ->Text = IniFile->ReadString("LatheTaperExt","FaceZ","-1.0");
	Cylinder->Checked = IniFile->ReadBool("LatheTaperExt","Cylinder",false);
	CylinderX->Text = IniFile->ReadString("LatheTaperExt","CylinderX","18.5");
	CylinderZ->Text = IniFile->ReadString("LatheTaperExt","CylinderZ","5");
	Feedrate->Text = IniFile->ReadString("LatheTaperExt","Feedrate","1");
	RoughingDOC->Text = IniFile->ReadString("LatheTaperExt","RoughingDOC","0.5");
	FinishDOC->Text = IniFile->ReadString("LatheTaperExt","FinishDOC","0.05");

	int numberoflines = IniFile->ReadInteger("LatheTaperExt","NumberofHeaderLines",0);
	String text;
	for (int i = 0; i < numberoflines; i++) {
		text = "HeaderLine" + String(i);
		Header->Lines->Add(IniFile->ReadString("LatheTaperExt",text," "));
		}
	numberoflines = IniFile->ReadInteger("LatheTaperExt","NumberofFooterLines",0);
	for (int i = 0; i < numberoflines; i++) {
		text = "FooterLine" + String(i);
		Footer->Lines->Add(IniFile->ReadString("LatheTaperExt",text," "));
		}

	delete IniFile;

	//format screen
	NumberLinesClick(Sender);
	WaypointClick(Sender);
	FacedClick(Sender);
	CylinderClick(Sender);

	//Update code generated
	UpdateGCode();
	//Now activate the OnChange events
	NumberLines->OnClick = NumberLinesClick;
	LineNumberIncrement->OnChange = UnitsClick;
	HideComments->OnClick = UnitsClick;
	Units->OnClick = UnitsClick;
	MeasurementMode->OnClick = UnitsClick;
	FeedRateMode->OnClick = UnitsClick;
    TaperComboBox->OnClick = UnitsClick;
	Waypoint->OnClick = WaypointClick;
	WaypointX->OnChange = UnitsClick;
	WaypointZ->OnChange = UnitsClick;
	StartpointX->OnChange = UnitsClick;
	StartpointZ->OnChange = UnitsClick;
	//EndpointX->OnChange = UnitsClick;
	//EndpointZ->OnChange = UnitsClick;
	ClearanceX->OnChange = ClearanceXChange;
	ClearanceZ->OnChange = UnitsClick;
	Faced->OnClick = FacedClick;
	FaceZ->OnChange = UnitsClick;
	Cylinder->OnClick = CylinderClick;
	CylinderX->OnChange = UnitsClick;
    CylinderZ->OnChange = UnitsClick;
	Feedrate->OnChange = UnitsClick;
	RoughingDOC->OnChange = UnitsClick;
	FinishDOC->OnChange = UnitsClick;
	Header->OnChange = UnitsClick;
	Footer->OnChange = UnitsClick;

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::NumberLinesClick(TObject *Sender)
{
	Label5->Enabled = NumberLines->Checked;
	LineNumberIncrement->Enabled = NumberLines->Checked;
	if (LineNumberIncrement->Enabled) {
		LineNumberIncrement->Color = clWindow;
		} else {
		LineNumberIncrement->Color = clBtnFace;
		}
	//Update code generated
	UpdateGCode();

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::WaypointClick(TObject *Sender)
{
	Label1->Enabled = Waypoint->Checked;
	WaypointX->Enabled = Waypoint->Checked;
	WaypointZ->Enabled = Waypoint->Checked;
	if (WaypointX->Enabled) {
		WaypointX->Color = clWindow;
		WaypointZ->Color = clWindow;
		} else {
		WaypointX->Color = clBtnFace;
		WaypointZ->Color = clBtnFace;
		}
	//Update code generated
	UpdateGCode();

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::FacedClick(TObject *Sender)
{
	Label12->Enabled = !Faced->Checked;
	FaceZ->Enabled = !Faced->Checked;
	if (!Faced->Checked) {
		FaceZ->Color = clWindow;
		} else {
		FaceZ->Color = clBtnFace;
        }
	//Update code generated
	UpdateGCode();
}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::CylinderClick(TObject *Sender)
{
	CylinderX->Enabled = Cylinder->Checked;
	CylinderZ->Enabled = Cylinder->Checked;
	if (CylinderX->Enabled) {
		CylinderX->Color = clWindow;
		CylinderZ->Color = clWindow;
		} else {
		CylinderX->Color = clBtnFace;
		CylinderZ->Color = clBtnFace;
		}
	//Update code generated
	UpdateGCode();

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::UnitsClick(TObject *Sender)
{
	//Update code generated
	UpdateGCode();

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::ClearanceXChange(TObject *Sender)
{
    ClearanceZ->Text = ClearanceX->Text;

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::FormDestroy(TObject *Sender)
{
	//Read settings & put them here
	TIniFile *IniFile = new TIniFile(ExtractFileDir(Application->ExeName) + "\\Data.ini");

	IniFile->EraseSection("LatheTaperExt");
	IniFile->WriteBool("LatheTaperExt","NumberLines",NumberLines->Checked);
	IniFile->WriteString("LatheTaperExt","LineNumberIncrement",LineNumberIncrement->Text);
	IniFile->WriteBool("LatheTaperExt","HideComments",HideComments->Checked);
	IniFile->WriteInteger("LatheTaperExt","Units",Units->ItemIndex);
	IniFile->WriteInteger("LatheTaperExt","MeasurementMode",MeasurementMode->ItemIndex);
	IniFile->WriteInteger("LatheTaperExt","FeedRateMode",FeedRateMode->ItemIndex);
	IniFile->WriteInteger("LatheTaperExt","TaperSize",TaperComboBox->ItemIndex);
	IniFile->WriteBool("LatheTaperExt","Waypoint",Waypoint->Checked);
	IniFile->WriteString("LatheTaperExt","WaypointX",WaypointX->Text);
	IniFile->WriteString("LatheTaperExt","WaypointZ",WaypointZ->Text);
	IniFile->WriteString("LatheTaperExt","StartpointX",StartpointX->Text);
	IniFile->WriteString("LatheTaperExt","StartpointZ",StartpointZ->Text);
	//IniFile->WriteString("LatheTaperExt","EndpointX",EndpointX->Text);
	//IniFile->WriteString("LatheTaperExt","EndpointZ",EndpointZ->Text);
	IniFile->WriteString("LatheTaperExt","ClearanceX",ClearanceX->Text);
	IniFile->WriteString("LatheTaperExt","ClearanceZ",ClearanceZ->Text);
	IniFile->WriteBool("LatheTaperExt","Faced",Faced->Checked);
	IniFile->WriteString("LatheTaperExt","FaceZ",FaceZ->Text);
	IniFile->WriteBool("LatheTaperExt","Cylinder",Cylinder->Checked);
	IniFile->WriteString("LatheTaperExt","CylinderX",CylinderX->Text);
	IniFile->WriteString("LatheTaperExt","CylinderZ",CylinderZ->Text);
	IniFile->WriteString("LatheTaperExt","Feedrate",Feedrate->Text);
	IniFile->WriteString("LatheTaperExt","RoughingDOC",RoughingDOC->Text);
	IniFile->WriteString("LatheTaperExt","FinishDOC",FinishDOC->Text);

	IniFile->WriteInteger("LatheTaperExt","NumberofHeaderLines",Header->Lines->Count);
	String text;
	for (int i = 0; i < Header->Lines->Count; i++) {
		text = "HeaderLine" + String(i);
		IniFile->WriteString("LatheTaperExt",text,Header->Lines->Strings[i]);
		}
	IniFile->WriteInteger("LatheTaperExt","NumberofFooterLines",Footer->Lines->Count);
	for (int i = 0; i < Footer->Lines->Count; i++) {
		text = "FooterLine" + String(i);
		IniFile->WriteString("LatheTaperExt",text,Footer->Lines->Strings[i]);
		}

	delete IniFile;

	Tapers.Length = 0;


}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::SaveAs1Click(TObject *Sender)
{
	//save code
	if (SaveDialog1->Execute()) {
        Memo1->Lines->SaveToFile(SaveDialog1->FileName);
		}

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::Exit1Click(TObject *Sender)
{
    Close();

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::Edit1Click(TObject *Sender)
{
	Copy1->Enabled = (Memo1->SelLength > 0);
    Paste1->Enabled = (Clipboard()->HasFormat(CF_TEXT));

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::SelectAll1Click(TObject *Sender)
{
	Memo1->SelStart = 0;
    Memo1->SelectAll();

}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::Paste1Click(TObject *Sender)
{
	Memo1->PasteFromClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::Copy1Click(TObject *Sender)
{
	Memo1->CopyToClipboard();
}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::Help2Click(TObject *Sender)
{
    //Open PDF help sheet
	String Location = ExtractFileDir(Application->ExeName) + "\\ExternalTaperInstructions.pdf";
	String action = "open";
	ShellExecute(Application->Handle,action.c_str(),Location.c_str(),Location.c_str(),NULL,SW_SHOW);
}
//---------------------------------------------------------------------------

void __fastcall TTaperForm::UpdateGCode(void)
{
	//read data
	Memo1->Lines->Clear();
	int linenumberincrement = LineNumberIncrement->Text.ToIntDef(1);
	if (linenumberincrement < 1) linenumberincrement = 1;
	int currentlinenumber = linenumberincrement;
	String text;

	double waypointX = ToDoubleDef(WaypointX->Text,0.0);
	double waypointZ = ToDoubleDef(WaypointZ->Text,0.0);
	double startpointX = ToDoubleDef(StartpointX->Text,0.0);
	double startpointZ = ToDoubleDef(StartpointZ->Text,0.0);
	taper Taper = Tapers[TaperComboBox->ItemIndex];
	if (MeasurementMode->ItemIndex != 0) {
		//change diameters to radii
		Taper.smalldia = Taper.smalldia/2.0;
		Taper.largedia = Taper.largedia/2.0;
		}
	if (Units->ItemIndex != 0) {
		//convert taper dimensions to inches
		Taper.taperlength = Taper.taperlength/25.4;
		Taper.smalldia = Taper.smalldia/25.4;
		Taper.largedia = Taper.largedia/25.4;
		}
	if (Taper.largedia > startpointX) {
        //This warning will always be displayed.
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "(WARNING. BAR DIAMETER TOO SMALL FOR TAPER SELECTED)";
		Memo1->Lines->Add(text);
		}
	double endpointX;// = ToDoubleDef(EndpointX->Text,0.0);
	double endpointZ;// = ToDoubleDef(EndpointZ->Text,0.0);
	double clearanceX = fabs(ToDoubleDef(ClearanceX->Text,1.0));
	double clearanceZ = fabs(ToDoubleDef(ClearanceZ->Text,1.0));
	double faceZ = ToDoubleDef(FaceZ->Text,0.0);
	if (faceZ > startpointZ) {
		faceZ = 2.0 * startpointZ - faceZ;
		}
	double cylinderX = fabs(ToDoubleDef(CylinderX->Text,0.0));
	if (cylinderX > Taper.smalldia) cylinderX = Taper.smalldia;
	double cylinderZ = fabs(ToDoubleDef(CylinderZ->Text,0.0));
	if (cylinderZ > (Taper.taperlength - 0.001)) cylinderZ = Taper.taperlength;
	double feedrate = ToDoubleDef(Feedrate->Text,1.0);
	double roughingdoc = fabs(ToDoubleDef(RoughingDOC->Text,0.5));
	double finishdoc = fabs(ToDoubleDef(FinishDOC->Text,0.05));
	double endoffset = finishdoc;
	if (!HideComments->Checked) {
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "(Taper to be Machined = " + Taper.Name + " )";
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "(Start Point X" + FormatFloat("####0.0000",startpointX) + " Z" + FormatFloat("####0.0000",startpointZ) + ")";
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "(Clearances X" + FormatFloat("####0.00##",clearanceX) + " Z" + FormatFloat("####0.00##",clearanceZ) + ")";
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "(Depth of Cut    Roughing:" + FormatFloat("####0.00##",roughingdoc) + "    Finish:" + FormatFloat("####0.00##",finishdoc) + ")";
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		if (Faced->Checked) {
			text += "(The Work is already faced)";
			} else {
			text += "(The Work is faced to Z = " + FormatFloat("####0.0000",faceZ) + ")";
			}
		Memo1->Lines->Add(text);
		if (Cylinder->Checked) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "(The Taper is reduced in length for a cylinder at the small end)";
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "(Cylinder size:  X" + FormatFloat("####0.0000",cylinderX) + "  Z" + FormatFloat("####0.0000",cylinderZ) + " )";
			Memo1->Lines->Add(text);
			}
		}




	//Add the Header text to the program file
	if (Header->Lines->Count > 0) {
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Header)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		for (int i = 0; i < Header->Lines->Count; i++) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += Header->Lines->Strings[i].UpperCase();
			Memo1->Lines->Add(text);
			}
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(End of Header)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		}

	text = "";
	if (NumberLines->Checked) {
		text = "N" + String(currentlinenumber) + " ";
		if (!HideComments->Checked) currentlinenumber += linenumberincrement;
		}
	text += "(Set Absolute Units, Set Units, Set Diameter/Radius Mode, Set Feedrate Mode)";
	if (!HideComments->Checked) Memo1->Lines->Add(text);
	text = "";
	if (NumberLines->Checked) {
		text = "N" + String(currentlinenumber) + " ";
		currentlinenumber += linenumberincrement;
		}
	text += "G90 ";
	if (Units->ItemIndex == 0) {
		text += "G21 ";
		} else {
		text += "G20 ";
		}
	if (MeasurementMode->ItemIndex == 0) {
		text += "G7 ";
		} else {
		text += "G8 ";
		}
	if (FeedRateMode->ItemIndex == 0) {
		text += "G94 ";
		} else {
		text += "G95 ";
		}
	Memo1->Lines->Add(text);


	//Waypoint
	if (Waypoint->Checked) {
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Rapid Move to waypoint to avoid work/tooling)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.0000",waypointX) + " Z" + FormatFloat("####0.0000",waypointZ);
		Memo1->Lines->Add(text);
		}


	text = "";
	if (NumberLines->Checked) {
		text = "N" + String(currentlinenumber) + " ";
		if (!HideComments->Checked) currentlinenumber += linenumberincrement;
		}
	text += "(Rapid Move to start point + X Clearance + Z Clearance)";
	if (!HideComments->Checked) Memo1->Lines->Add(text);
	text = "";
	if (NumberLines->Checked) {
		text = "N" + String(currentlinenumber) + " ";
		currentlinenumber += linenumberincrement;
		}
	text += "G00 X" + FormatFloat("####0.0000",startpointX + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
	Memo1->Lines->Add(text);

	//set the feedrate
	text = "";
	if (NumberLines->Checked) {
		text = "N" + String(currentlinenumber) + " ";
		currentlinenumber += linenumberincrement;
		}
	text += "F" + FormatFloat("####0.####",feedrate);
	Memo1->Lines->Add(text);


	double lastroughingZ;
	int numberofloops;
	double lastroughingZpos;
	double lastroughingdoc;
	bool lastrough;


	//Face end
	if (!Faced->Checked) {
		lastroughingZ = faceZ + finishdoc;
		numberofloops = (int)((startpointZ - lastroughingZ)/roughingdoc);
		lastroughingZpos = (startpointZ - ((double)(numberofloops))*roughingdoc);
		lastroughingdoc = lastroughingZpos - lastroughingZ;
		lastrough = true;
		if (lastroughingdoc == 0.0) lastrough = false;
		if (lastroughingdoc == roughingdoc) {
			numberofloops++;
			lastrough = false;
			}
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Facing End of work while at max diameter for max stiffness)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Facing goes " + FormatFloat("####0.0000",finishdoc) + " past Lathe CL to cut off centre pip)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);

		for (int i = 0; i < numberofloops + 1; i++) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				if (!HideComments->Checked) currentlinenumber += linenumberincrement;
				}
			text += "(Face Roughing Loop " + String(i+1) + " )";
			if (!HideComments->Checked) Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(i)) * roughingdoc);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",-finishdoc) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(i)) * roughingdoc);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000", + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(i)) * roughingdoc + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(i)) * roughingdoc + clearanceZ);
			Memo1->Lines->Add(text);
			}
		if (lastrough) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				if (!HideComments->Checked) currentlinenumber += linenumberincrement;
				}
			text += "(Face Roughing Loop " + String(numberofloops + 2) + " )";
			if (!HideComments->Checked) Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX  + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(numberofloops)) * roughingdoc - lastroughingdoc);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",-finishdoc ) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(numberofloops)) * roughingdoc - lastroughingdoc);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000", clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ - ((double)(numberofloops)) * roughingdoc - lastroughingdoc + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX  + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ  - ((double)(numberofloops)) * roughingdoc - lastroughingdoc + clearanceZ);
			Memo1->Lines->Add(text);
			}

		//finishing cut
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Face Finishing Cut)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.0000",startpointX  + clearanceX) + " Z" + FormatFloat("####0.0000",faceZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.0000",-finishdoc) + " Z" + FormatFloat("####0.0000",faceZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.0000",clearanceX) + " Z" + FormatFloat("####0.0000",faceZ + clearanceZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.0000",startpointX) + " Z" + FormatFloat("####0.0000",faceZ + clearanceZ);
		Memo1->Lines->Add(text);

		//reset the startpointZ as faceZ so I can re-use the cylinder code
		startpointZ = faceZ;
		}
	//OK. The work has now been faced off.

	double lastroughingX;
	double lastroughingdia;

	if (Cylinder->Checked) {
		//this code copied wholesale from cylinder turning program
		endpointX = cylinderX;
		endpointZ = startpointZ - cylinderZ;
		//effect the change in the taper small end and length
		taper taperold = Taper;
		Taper.taperlength = Taper.taperlength - cylinderZ;
		Taper.smalldia = Taper.largedia - (Taper.largedia - Taper.smalldia) * Taper.taperlength / taperold.taperlength;

		//now to do the cylinder turning
		lastroughingX = endpointX + finishdoc;
		numberofloops = (int)((startpointX - lastroughingX)/roughingdoc);
		lastroughingdia = (startpointX - ((double)(numberofloops))*roughingdoc);
		lastroughingdoc = lastroughingdia - lastroughingX;
		lastrough = true;
		if (lastroughingdoc == 0.0) lastrough = false;
		if (lastroughingdoc == roughingdoc) {
			numberofloops++;
			lastrough = false;
			}
		for (int i = 0; i < numberofloops + 1; i++) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				if (!HideComments->Checked) currentlinenumber += linenumberincrement;
				}
			text += "(Cylinder Roughing Loop " + String(i+1) + " )";
			if (!HideComments->Checked) Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * roughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * roughingdoc) + " Z" + FormatFloat("####0.0000",endpointZ + endoffset);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * roughingdoc + clearanceX) + " Z" + FormatFloat("####0.0000",endpointZ + finishdoc + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * roughingdoc + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			}
		if (lastrough) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				if (!HideComments->Checked) currentlinenumber += linenumberincrement;
				}
			text += "(Cylinder Roughing Loop " + String(numberofloops + 2) + " )";
			if (!HideComments->Checked) Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * roughingdoc - lastroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * roughingdoc - lastroughingdoc) + " Z" + FormatFloat("####0.0000",endpointZ + endoffset);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * roughingdoc - lastroughingdoc + clearanceX) + " Z" + FormatFloat("####0.0000",endpointZ + finishdoc + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * roughingdoc - lastroughingdoc + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			}

		//finishing cut
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Cylinder Finishing Cut)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.0000",endpointX) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.0000",endpointX) + " Z" + FormatFloat("####0.0000",endpointZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.0000",startpointX + clearanceX) + " Z" + FormatFloat("####0.0000",endpointZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.0000",startpointX ) + " Z" + FormatFloat("####0.0000",endpointZ + clearanceZ);
		Memo1->Lines->Add(text);

		//move the start point onto the beginning of the taper.
		startpointZ = endpointZ;
		}

	//now do the taper.
	double taperZ = Taper.taperlength;
	double taperdeltaX = Taper.largedia - Taper.smalldia;
	double taperratio = taperdeltaX / taperZ;
	double angle = acosD(taperratio);
	double bardiameter = startpointX;
	if (MeasurementMode->ItemIndex == 0) {
		angle = angle/2.0 ;
		}
	double angleroughingdoc = roughingdoc/cosD(angle);
	double anglefinishdoc = finishdoc/cosD(angle);

	//endpointZ = startpointZ - taperZ + finishdoc;
	startpointX = startpointX - angleroughingdoc;
	if (startpointX < (Taper.smalldia + anglefinishdoc)) {
		startpointX = (Taper.smalldia + anglefinishdoc);
		}


	lastroughingX = Taper.smalldia + anglefinishdoc;
	numberofloops = (int)((startpointX - lastroughingX)/angleroughingdoc);
	lastroughingdia = (startpointX - ((double)(numberofloops))*angleroughingdoc);
	lastroughingdoc = lastroughingdia - lastroughingX;
	lastrough = true;
	if (lastroughingdoc == 0.0) lastrough = false;
	if (lastroughingdoc == angleroughingdoc) {
		numberofloops++;
		lastrough = false;
		}

	for (int i = 0; i < numberofloops + 1; i++) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				if (!HideComments->Checked) currentlinenumber += linenumberincrement;
				}
			text += "(Taper Roughing Loop " + String(i+1) + " )";
			if (!HideComments->Checked) Memo1->Lines->Add(text);
			//rapid to same dia as cut start
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * angleroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			//feed to cut start
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * angleroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ);
			Memo1->Lines->Add(text);
			//feed to end of taper
			//first find end of taper
			endpointZ = startpointZ - taperZ + finishdoc;
			endpointX = startpointX  - (((double)(i)) * angleroughingdoc) + taperratio * taperZ;
			if (endpointX > bardiameter) {
				//OK, haven't reached the shoulder at the end of the taper.
				endpointX = bardiameter;
				endpointZ = startpointZ - (startpointX - (startpointX  - (((double)(i)) * angleroughingdoc)))/taperratio;
				}
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",endpointX) + " Z" + FormatFloat("####0.0000",endpointZ);
			Memo1->Lines->Add(text);
			//rapid away from shoulder
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",endpointX + clearanceX) + " Z" + FormatFloat("####0.0000",endpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * angleroughingdoc + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			}
		if (lastrough) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				if (!HideComments->Checked) currentlinenumber += linenumberincrement;
				}
			text += "(Taper Roughing Loop " + String(numberofloops + 2) + " )";
			if (!HideComments->Checked) Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			//text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * angleroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * angleroughingdoc - lastroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			//text += "G01 X" + FormatFloat("####0.0000",startpointX - ((double)(i)) * angleroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ);
			text += "G01 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * angleroughingdoc - lastroughingdoc) + " Z" + FormatFloat("####0.0000",startpointZ);
			Memo1->Lines->Add(text);
			//feed to end of taper
			//first find end of taper
			endpointZ = startpointZ - taperZ + finishdoc;
			endpointX = startpointX  - (((double)(numberofloops)) * angleroughingdoc) + taperratio * taperZ;
			if (endpointX > bardiameter) {
				//OK, haven't reached the shoulder at the end of the taper.
				endpointX = bardiameter;
				endpointZ = startpointZ - ((startpointX - (startpointX  - ((double)(numberofloops)) * angleroughingdoc)))/taperratio;
				}
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G01 X" + FormatFloat("####0.0000",endpointX) + " Z" + FormatFloat("####0.0000",endpointZ);
			Memo1->Lines->Add(text);
			//rapid away from shoulder
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			//text += "G00 X" + FormatFloat("####0.0000",endpointX + clearanceX) + " Z" + FormatFloat("####0.0000",endpointZ + clearanceZ);
			text += "G00 X" + FormatFloat("####0.0000",endpointX + clearanceX) + " Z" + FormatFloat("####0.0000",endpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += "G00 X" + FormatFloat("####0.0000",startpointX - ((double)(numberofloops)) * roughingdoc - lastroughingdoc + clearanceX) + " Z" + FormatFloat("####0.0000",startpointZ + clearanceZ);
			Memo1->Lines->Add(text);
			}

		//finishing cut
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Taper Finishing Cut)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.00000",Taper.smalldia) + " Z" + FormatFloat("####0.00000",startpointZ + clearanceZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.00000",Taper.smalldia) + " Z" + FormatFloat("####0.00000",startpointZ);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.00000",Taper.largedia) + " Z" + FormatFloat("####0.00000",startpointZ - Taper.taperlength);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G01 X" + FormatFloat("####0.00000",bardiameter + clearanceX) + " Z" + FormatFloat("####0.00000",startpointZ - Taper.taperlength);
		Memo1->Lines->Add(text);
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			currentlinenumber += linenumberincrement;
			}
		text += "G00 X" + FormatFloat("####0.0000",bardiameter + clearanceX ) + " Z" + FormatFloat("####0.0000",startpointZ - Taper.taperlength + clearanceZ);
		Memo1->Lines->Add(text);




	//Add the Footer text to the program file
	if (Footer->Lines->Count > 0) {
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(Footer)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		for (int i = 0; i < Footer->Lines->Count; i++) {
			text = "";
			if (NumberLines->Checked) {
				text = "N" + String(currentlinenumber) + " ";
				currentlinenumber += linenumberincrement;
				}
			text += Footer->Lines->Strings[i].UpperCase();
			Memo1->Lines->Add(text);
			}
		text = "";
		if (NumberLines->Checked) {
			text = "N" + String(currentlinenumber) + " ";
			if (!HideComments->Checked) currentlinenumber += linenumberincrement;
			}
		text += "(End of Footer)";
		if (!HideComments->Checked) Memo1->Lines->Add(text);
		}

	//Return focus to top of memo list
	Memo1->SelStart = 0;
	Memo1->SelLength = 1;
	Memo1->SelLength = 0;

}
//---------------------------------------------------------------------------

